import type { UserInDatabase } from "@interfaces/db/mariadb.interface";
import type { IRegisterInput } from "@interfaces/services/register.types";
import JwtService from "@services/authentication/jwt.service";
import RegisterService from "@services/authentication/register.service";
import UsersService from "@services/users.service";
import ConvertersUtils from "@utils/converters.util";
import { HeaderUtil } from "@utils/headers.util";
import { LogsUtils } from "@utils/logs.util";
import ValidatorsUtils from "@utils/validators.util";
import { HttpStatusCode } from "axios";
import type { Request, Response } from "express";
import type { JWTPayload } from "jose";

const logs = new LogsUtils("AuthController");

async function registerController(req: Request, res: Response) {
	const body = req.body;
	if (!body.username || !body.email || !body.password) {
		logs.warn("Missing required fields", req.ip);
		return res
			.status(HttpStatusCode.BadRequest)
			.json({ error: "Missing required fields" });
	}
	if (!ValidatorsUtils.isEmail(body.email)) {
		logs.warn("Invalid email format", req.ip);
		return res
			.status(HttpStatusCode.BadRequest)
			.json({ error: "Invalid email format" });
	}
	if (!ValidatorsUtils.isUsername(body.username)) {
		logs.warn("Invalid username format", req.ip);
		return res
			.status(HttpStatusCode.BadRequest)
			.json({ error: "Invalid username format" });
	}
	if (!ValidatorsUtils.isPassword(body.password)) {
		logs.warn("Invalid password format", req.ip);
		return res
			.status(HttpStatusCode.BadRequest)
			.json({ error: "Invalid password format" });
	}

	const UserIfExist = {
		byEmail: await UsersService.get.byEmail(body.email),
		byUsername: await UsersService.get.byUsername(body.username),
	};

	if (UserIfExist.byEmail.length > 0 || UserIfExist.byUsername.length > 0) {
		logs.warn("User already exists", req.ip);
		return res
			.status(HttpStatusCode.Found)
			.json({ error: "User already exists" });
	}

	const data: IRegisterInput = {
		username: body.username,
		email: body.email,
		password: body.password,
	};
	if (body.displayName) data.displayName = body.displayName;

	const registerResult = await RegisterService(data);

	if (!registerResult.success) {
		logs.error(registerResult.message, req.ip);
		return res
			.status(HttpStatusCode.InternalServerError)
			.json({ error: registerResult.message });
	}

	//TODO Image handling

	logs.info("User registered successfully", req.ip);
	return res.status(HttpStatusCode.Created).json({
		message: "User registered successfully",
		token: registerResult.token,
		id: registerResult.id,
	});
}

async function GetSelfController(req: Request, res: Response) {
	const originToken = (await HeaderUtil.getToken(req)) as string;
	const token = (await JwtService.verify(originToken)) as JWTPayload;
	const userInDatabases = (await UsersService.get.byId(
		token.sub as string,
	)) as Array<UserInDatabase>;
	if (userInDatabases.length === 0 || !userInDatabases[0]) {
		logs.warn("User not found", req.ip);
		return res.status(HttpStatusCode.Gone).json({ error: "User not found" });
	}
	const User = userInDatabases[0];
	return res.status(HttpStatusCode.Found).json({
		id: User.id,
		username: User.username,
		email: User.email,
		displayName: User.display_name,
		isAdmin: ConvertersUtils.bufferToBool(User.admin as Buffer),
		followers: await UsersService.get.followers(User.id),
		following: await UsersService.get.following(User.id),
	});
}

const AuthController = {
	register: registerController,
	me: GetSelfController,
};

export default AuthController;
