import type {
	FollowInDatabase,
	UserInDatabase,
} from "@interfaces/db/mariadb.interface";
import type { IComment, IPost } from "@interfaces/db/mongodb.types";
import { MariadbService } from "@services/databases/mariadb.service";
import { MongodbService } from "@services/databases/mongodb.service";
import { LogsUtils } from "@utils/logs.util";

interface MariaDbStatusResult {
	fieldCount: number;
	affectedRows: number;
	insertId: number;
	info: string;
	serverStatus: number;
	warningStatus: number;
	changedRows: number;
}

export class DatabasesService {
	private readonly _appName;
	private readonly maria: MariadbService;
	private readonly mongo: MongodbService;
	private readonly logs: LogsUtils;
	constructor(appName?: string) {
		this._appName = appName || "App";
		this.maria = new MariadbService(`${this._appName} DbInteractions`);
		this.mongo = new MongodbService(`${this._appName} DbInteractions`);
		this.logs = new LogsUtils(`${this._appName} DbInteractions`);
	}

	async getAllUsers() {
		const result = (await this.maria.query(
			"SELECT * FROM users",
		)) as unknown as Array<UserInDatabase>;
		this.logs.debug(
			"Fetching users from database...",
			`${result?.length} user(s) found.`,
		);
		return result;
	}

	async getUserById(id: string) {
		const result = (await this.maria.execute(
			"SELECT * FROM users WHERE id = ?",
			[id],
		)) as unknown as Array<UserInDatabase>;
		this.logs.debug(
			`Fetching user with id ${id} from database...`,
			`${result?.length} user(s) found.`,
		);
		return result;
	}

	async getUserByUsername(username: string) {
		const result = (await this.maria.execute(
			"SELECT * FROM users WHERE username = ?",
			[username],
		)) as unknown as Array<UserInDatabase>;
		this.logs.debug(
			`Fetching user with username "${username}" from database...`,
			`${result?.length} user(s) found.`,
		);
		return result;
	}

	async getUserByEmail(email: string) {
		const result = (await this.maria.execute(
			"SELECT * FROM users WHERE email = ?",
			[email],
		)) as unknown as Array<UserInDatabase>;
		this.logs.debug(
			`Fetching user with email "${email}" from database...`,
			`${result?.length} user(s) found.`,
		);
		return result;
	}

	async getFollowingById(id: string) {
		const result = (await this.maria.execute(
			"SELECT * FROM follows WHERE source_id = ?",
			[id],
		)) as unknown as Array<FollowInDatabase>;
		this.logs.debug(
			`Fetching followed users for user with id ${id} from database...`,
			`${result?.length} user(s) found.`,
		);
		return result;
	}

	async getFollowersById(id: string) {
		const result = (await this.maria.execute(
			"SELECT * FROM follows WHERE target_id = ?",
			[id],
		)) as unknown as Array<FollowInDatabase>;
		this.logs.debug(
			`Fetching followers for user with id ${id} from database...`,
			`${result?.length} user(s) found.`,
		);
		return result;
	}

	async insertUser(user: UserInDatabase): Promise<boolean> {
		const factorized = await this.maria.factorize({
			values: user,
			actionName: "Inserting a user",
			throwOnError: true,
		});
		//TODO if no id stop
		const valuesArray = factorized._valuesArray;
		const questionMarks = factorized._questionMarksFields;
		const keysArray = factorized._keysTemplate;
		const _sql = `INSERT INTO users (${keysArray}) VALUES (${questionMarks})`;
		try {
			const result = (await this.maria.execute(
				_sql,
				valuesArray,
			)) as unknown as MariaDbStatusResult;
			this.logs.debug(
				`Inserted new user with id ${user.id}`,
				`Rows affected: ${result.affectedRows}`,
			);
			return true;
		} catch (err) {
			this.logs.softError("An error occurred.", err);
			return false;
		}
	}
	//ToTest
	async editUser(userId: string, data: object): Promise<object> {
		const factorized = await this.maria.factorize({
			values: data,
			actionName: "Editing a user",
			throwOnError: true,
		});
		const valuesArray = factorized._valuesArray;
		const keysArray = factorized._keysTemplate.split(",");
		const setFields = keysArray.map((key) => `${key} = ?`).join(", ");
		const _sql = `UPDATE users SET ${setFields} WHERE id = ?`;
		valuesArray.push(userId);
		try {
			const result = (await this.maria.execute(
				_sql,
				valuesArray,
			)) as unknown as MariaDbStatusResult;
			this.logs.debug(
				`Edited user with id ${userId}`,
				`Rows affected: ${result.affectedRows}`,
			);
			return result;
		} catch (err) {
			this.logs.softError("An error occurred.", err);
			return {};
		}
	}
	//ToTest
	async createFollow(sourceId: string, targetId: string): Promise<boolean> {
		const _sql = "INSERT INTO follows (source_id, target_id) VALUES (?, ?)";
		try {
			const result = (await this.maria.execute(_sql, [
				sourceId,
				targetId,
			])) as unknown as MariaDbStatusResult;
			this.logs.debug(
				`Created follow relationship between user with id ${sourceId} and user with id ${targetId}`,
				`Rows affected: ${result.affectedRows}`,
			);
			return true;
		} catch (err) {
			this.logs.softError("An error occurred.", err);
			return false;
		}
	}
	//ToTest
	async deleteFollow(sourceId: string, targetId: string): Promise<boolean> {
		const _sql = "DELETE FROM follows WHERE source_id = ? AND target_id = ?";
		try {
			const result = (await this.maria.execute(_sql, [
				sourceId,
				targetId,
			])) as unknown as MariaDbStatusResult;
			this.logs.debug(
				`Deleted follow relationship between user with id ${sourceId} and user with id ${targetId}`,
				`Rows affected: ${result.affectedRows}`,
			);
			return true;
		} catch (err) {
			this.logs.softError("An error occurred.", err);
			return false;
		}
	}
	//ToTest
	async deleteUser(userId: string): Promise<boolean> {
		const _sql = "DELETE FROM users WHERE id = ?";
		try {
			const result = (await this.maria.execute(_sql, [
				userId,
			])) as unknown as MariaDbStatusResult;
			this.logs.debug(
				`Deleted user with id ${userId}`,
				`Rows affected: ${result.affectedRows}`,
			);
			return true;
		} catch (err) {
			this.logs.softError("An error occurred.", err);
			return false;
		}
	}
	//ToTest
	async createPost(data: IPost): Promise<string | boolean> {
		try {
			const result = await this.mongo.use().collection("posts").insertOne(data);
			if (result.acknowledged) {
				return data.id;
			}
			return false;
		} catch (err) {
			this.logs.softError("Error when inserting a post.", err);
			return false;
		}
	}
	//ToTest
	async getPostById(id: string): Promise<IPost | false> {
		try {
			const result = (await this.mongo
				.use()
				.collection("posts")
				.findOne({ id })) as unknown as IPost;
			if (!result) return false;
			return result;
		} catch (err) {
			this.logs.softError("Error when fetching a post.", err);
			return false;
		}
	}
	//ToTest
	async getPostsByUserId(userId: string): Promise<IPost[]> {
		try {
			const result = (await this.mongo
				.use()
				.collection("posts")
				.find({ userId })
				.toArray()) as unknown as IPost[];
			return result;
		} catch (err) {
			this.logs.softError("Error when fetching posts by user id.", err);
			return [];
		}
	}
	//ToTest
	async createComment(data: IComment): Promise<string | boolean> {
		try {
			const result = await this.mongo
				.use()
				.collection("comments")
				.insertOne(data);
			if (result.acknowledged) {
				return data.id;
			}
			return false;
		} catch (err) {
			this.logs.softError("Error when inserting a comment.", err);
			return false;
		}
	}
	//ToTest
	async getCommentsByPostId(postId: string): Promise<IComment[]> {
		try {
			const result = (await this.mongo
				.use()
				.collection("comments")
				.find({ postId })
				.toArray()) as unknown as IComment[];
			return result;
		} catch (err) {
			this.logs.softError("Error when fetching comments by post id.", err);
			return [];
		}
	}
	//ToTest
	async getPostsWithCommentsByUserId(
		userId: string,
	): Promise<Array<{ post: IPost; comments: IComment[] | undefined }>> {
		try {
			const posts = await this.getPostsByUserId(userId);
			const postIds = posts.map((post) => post.id);
			const comments = await Promise.all(
				postIds.map((postId) => this.getCommentsByPostId(postId)),
			);
			const result = posts.map((post, index) => ({
				post,
				comments: comments[index],
			}));
			this.logs.debug(
				"Posts with comments by user id:",
				result.map((item) => ({
					postId: item.post.id,
					commentsCount: item.comments?.length || 0,
				})),
			);

			return result;
		} catch (err) {
			this.logs.softError(
				"Error when fetching posts with comments by user id.",
				err,
			);
			return [];
		}
	}
	//ToTest
	async deleteComment(commentId: string): Promise<boolean> {
		try {
			const result = await this.mongo
				.use()
				.collection("comments")
				.deleteOne({ id: commentId });
			this.logs.debug(
				`Deleted comment with id ${commentId}`,
				`Rows affected: ${result.deletedCount}`,
			);
			return true;
		} catch (err) {
			this.logs.softError("An error occurred.", err);
			return false;
		}
	}
	//ToTest
	async deletePost(postId: string): Promise<boolean> {
		try {
			const result = await this.mongo
				.use()
				.collection("posts")
				.deleteOne({ id: postId });
			this.logs.debug(
				`Deleted post with id ${postId}`,
				`Rows affected: ${result.deletedCount}`,
			);
			return true;
		} catch (err) {
			this.logs.softError("An error occurred.", err);
			return false;
		}
	}
	//ToTest
	async getFollowersCountById(id: string) {
		const result = (await this.maria.execute(
			"SELECT COUNT(*) FROM follows WHERE target_id = ?",
			[id],
		)) as unknown as number;
		this.logs.debug(
			`Fetching followers count for user with id ${id} from database...`,
			`${result} follower(s) found.`,
		);
		return result;
	}
	//ToTest
	async getFollowingCountById(id: string) {
		const result = (await this.maria.execute(
			"SELECT COUNT(*) AS count FROM follows WHERE source_id = ?",
			[id],
		)) as unknown as Array<{ count: number }>;
		this.logs.debug(
			`Fetching count of followed users for user with id ${id} from database...`,
			`${result?.[0]?.count} user(s) found.`,
		);
		return result?.[0]?.count || 0;
	}
	//ToTest
	async getMostFollowedUser() {
		const _sql =
			"SELECT target_id, COUNT(*) AS count FROM follows GROUP BY target_id ORDER BY count DESC LIMIT 3";
		const result = (await this.maria.query(_sql)) as unknown as Array<{
			target_id: string;
			count: number;
		}>;
		this.logs.debug(
			"Fetching most followed users from database...",
			`${result?.length} user(s) found.`,
		);
		return result;
	}

	async updatePost(data: IPost) {
		//MongoDB
		await this.mongo
			.use()
			.collection("posts")
			.updateOne({ id: data.id }, { $set: data });
	}
}

export const justForTesting = new DatabasesService("OnlyDevs");
