import type {
	ISqlFactorizeInput,
	ISqlFactorizeOutput,
} from "@interfaces/sqlfactorizer.interface";
import { EnvUtils } from "@utils/env.util";
import { LogsUtils } from "@utils/logs.util";
import mysql, { type Connection } from "mysql2";

export class MariadbService {
	private envs: EnvUtils;
	private logs: LogsUtils;
	private readonly contextName: string;
	private Connection: Connection;

	constructor(contextName?: string) {
		this.envs = new EnvUtils(`MariaDB >> ${contextName}`);
		this.logs = new LogsUtils(`MariaDB >> ${contextName}`);
		this.contextName = contextName || "Unknown";
		this.logs = new LogsUtils(this.contextName);
		this.Connection = mysql.createConnection({
			host: `${this.envs.get("MYSQL_HOST")}`,
			port: Number.parseInt(`${this.envs.get("MYSQL_PORT")}`),
			user: `${this.envs.get("MYSQL_USERNAME")}`,
			database: `${this.envs.get("MYSQL_DATABASE")}`,
			password: `${this.envs.get("MYSQL_PASSWORD")}`,
		});
		this.Connection.connect((err) => {
			if (err) {
				this.logs.error("Error connecting to MySQL:", err);
			}
			this.logs.debug("Connected to MariaDB", this.envs.get("MYSQL_DATABASE"));
		});
	}
	closeConnection() {
		this.Connection.end();
	}

	query(queryString: string) {
		return new Promise((resolve, reject) => {
			this.Connection.query(queryString, (err, results) => {
				if (err) {
					this.logs.error("Error executing query:", err);
					reject(err);
				} else {
					resolve(results);
				}
			});
		});
	}

	/**
	 * Factorize the input data values into a database query.
	 * `id` field will not be injected in result to avoid problems.
	 *
	 * @param {ISqlFactorizeInput} data - The input data containing values to factorize.
	 * @return {Promise<ISqlFactorizeOutput>} - A promise resolving to the factorized output.
	 */
	factorize(data: ISqlFactorizeInput): Promise<ISqlFactorizeOutput> {
		return new Promise((resolve, reject) => {
			try {
				console.log(data);
				let _id = "";
				// @ts-ignore
				if (data.values.id) {
					// @ts-ignore
					_id = data.values.id;
					// @ts-ignore
					delete data.values.id;
				}
				const _sqlQueryKeys = Object.keys(data.values).map(
					(key: string) => `${key}`,
				);
				const values = Object.values(data.values).map((val) => val.toString());
				this.logs.debug(
					`Factorized ${_sqlQueryKeys.length} keys for a prepare Query.`,
					`Action: ${data.actionName}`,
				);
				if (_id.length > 0) {
					this.logs.trace("Id post-pushed in factorized data.", _id);
					values.push(_id);
					_sqlQueryKeys.push("id");
				}
				const sqlQueryKeys = _sqlQueryKeys.join(", ");

				const questionMarksFields = Array(values.length)
					.fill("?")
					.join(", ")
					.toString();

				const factorizedOutput: ISqlFactorizeOutput = {
					_keysTemplate: sqlQueryKeys,
					_questionMarksFields: questionMarksFields,
					totalFields: _sqlQueryKeys.length,
					_valuesArray: values,
				};
				resolve(factorizedOutput);
			} catch (err) {
				if (data.throwOnError) throw new Error(`${err}`);
				this.logs.softError(err);
				reject(`${err}`);
			}
		});
	}

	/**
	 * Executes a query using the provided queryString and values.
	 * @param {string} queryString - The SQL query string to execute.
	 * @param {Array<string | boolean | Date | number>} values - The values to be inserted into the query.
	 * @returns {Promise<unknown>} - A promise that resolves with the query results or rejects with an error.
	 */
	execute(
		queryString: string,
		values: Array<string | boolean | Date | number>,
	): Promise<unknown> {
		return new Promise((resolve, reject) => {
			this.Connection.execute(
				queryString,
				values,
				(err: mysql.QueryError | null, results: mysql.QueryResult) => {
					if (err) {
						reject(err);
					}
					resolve(results);
				},
			);
		});
	}

	/**
	 * Unprepare a previously prepared SQL query.
	 *
	 * @param {string} queryString
	 * 			- The SQL query string to unprepare.
	 * @return {Promise}
	 * 			- A promise that resolves if the unprepare operation is successful,
	 *        or rejects with an error if there was an error unpreparing the query.
	 */
	unprepare(queryString: string): Promise<unknown> {
		return new Promise((resolve, reject) => {
			try {
				resolve(this.Connection.unprepare(queryString));
			} catch (err) {
				reject(err);
			}
		});
	}
}
